import sys
from typing import TypeVar

from result import Err, Ok, Result
from textual.validation import ValidationResult, Validator

from sksmithy._parsers import name_parser, params_parser

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self

T = TypeVar("T")
R = TypeVar("R")


class _BaseValidator(Validator):
    @staticmethod
    def parser(value: str) -> Result[str | list[str], str]:  # pragma: no cover
        raise NotImplementedError

    def validate(self: Self, value: str) -> ValidationResult:
        match self.parser(value):
            case Ok(_):
                return self.success()
            case Err(msg):
                return self.failure(msg)


class NameValidator(_BaseValidator):
    @staticmethod
    def parser(value: str) -> Result[str, str]:
        return name_parser(value)


class ParamsValidator(_BaseValidator):
    @staticmethod
    def parser(value: str) -> Result[list[str], str]:
        return params_parser(value)



import re
import time
from importlib import resources
from importlib.metadata import version

from result import Err, Ok

from sksmithy._models import EstimatorType
from sksmithy._parsers import check_duplicates, name_parser, params_parser
from sksmithy._prompts import (
    PROMPT_DECISION_FUNCTION,
    PROMPT_ESTIMATOR,
    PROMPT_LINEAR,
    PROMPT_NAME,
    PROMPT_OPTIONAL,
    PROMPT_PREDICT_PROBA,
    PROMPT_REQUIRED,
    PROMPT_SAMPLE_WEIGHT,
)
from sksmithy._utils import render_template

if (st_version := version("streamlit")) and tuple(int(re.sub(r"\D", "", str(v))) for v in st_version.split(".")) < (
    1,
    34,
    0,
):  # pragma: no cover
    st_import_err_msg = (
        f"streamlit>=1.34.0 is required for this module. Found version {st_version}.\nInstall it with "
        '`python -m pip install "streamlit>=1.34.0"` or `python -m pip install "sklearn-smithy[streamlit]"`'
    )
    raise ImportError(st_import_err_msg)

else:  # pragma: no cover
    import streamlit as st

SIDEBAR_MSG: str = (resources.files("sksmithy") / "_static" / "description.md").read_text()


def app() -> None:  # noqa: C901,PLR0912,PLR0915
    """Streamlit App."""
    st.set_page_config(
        page_title="Smithy",
        page_icon="⚒️",
        layout="wide",
        menu_items={
            "Get Help": "https://github.com/FBruzzesi/sklearn-smithy",
            "Report a bug": "https://github.com/FBruzzesi/sklearn-smithy/issues/new",
            "About": """
                Forge your own scikit-learn estimator!

                For more information, please visit the [sklearn-smithy](https://github.com/FBruzzesi/sklearn-smithy)
                repository.
                """,
        },
    )

    st.title("Scikit-learn Smithy ⚒️")
    st.markdown("## Forge your own scikit-learn compatible estimator")

    with st.sidebar:
        st.markdown(SIDEBAR_MSG)

    linear = False
    predict_proba = False
    decision_function = False
    estimator_type: EstimatorType | None = None

    required_is_valid = False
    optional_is_valid = False
    msg_duplicated_params: str | None = None

    if "forged_template" not in st.session_state:
        st.session_state["forged_template"] = ""

    if "forge_counter" not in st.session_state:
        st.session_state["forge_counter"] = 0

    with st.container():  # name and type
        c11, c12 = st.columns(2)

        with c11:  # name
            name_input = st.text_input(
                label=PROMPT_NAME,
                value="MightyEstimator",
                placeholder="MightyEstimator",
                help=(
                    "It should be a valid "
                    "[python identifier](https://docs.python.org/3/reference/lexical_analysis.html#identifiers)"
                ),
                key="name",
            )

            match name_parser(name_input):
                case Ok(name):
                    pass
                case Err(name_error_msg):
                    name = ""
                    st.error(name_error_msg)

        with c12:  # type
            estimator = st.selectbox(
                label=PROMPT_ESTIMATOR,
                options=tuple(e.value for e in EstimatorType),
                format_func=lambda v: " ".join(x.capitalize() for x in v.split("-")),
                index=None,
                key="estimator",
            )

            if estimator:
                estimator_type = EstimatorType(estimator)

    with st.container():  # params
        c21, c22 = st.columns(2)

        with c21:  # required
            required_params = st.text_input(
                label=PROMPT_REQUIRED,
                placeholder="alpha,beta",
                help=(
                    "It should be a sequence of comma-separated "
                    "[python identifiers](https://docs.python.org/3/reference/lexical_analysis.html#identifiers)"
                ),
                key="required",
            )

            match params_parser(required_params):
                case Ok(required):
                    required_is_valid = True
                case Err(required_err_msg):
                    required_is_valid = False
                    st.error(required_err_msg)

        with c22:  # optional
            optional_params = st.text_input(
                label=PROMPT_OPTIONAL,
                placeholder="mu,sigma",
                help=(
                    "It should be a sequence of comma-separated "
                    "[python identifiers](https://docs.python.org/3/reference/lexical_analysis.html#identifiers)"
                ),
                key="optional",
            )

            match params_parser(optional_params):
                case Ok(optional):
                    optional_is_valid = True
                case Err(optional_err_msg):
                    optional_is_valid = False
                    st.error(optional_err_msg)

        if required_is_valid and optional_is_valid and (msg_duplicated_params := check_duplicates(required, optional)):
            st.error(msg_duplicated_params)

    with st.container():  # sample_weight and linear
        c31, c32 = st.columns(2)

        with c31:  # sample_weight
            sample_weight = st.toggle(
                PROMPT_SAMPLE_WEIGHT,
                help="[sample_weight](https://scikit-learn.org/dev/glossary.html#term-sample_weight)",
                key="sample_weight",
            )
        with c32:  # linear
            linear = st.toggle(
                label=PROMPT_LINEAR,
                disabled=(estimator_type not in {EstimatorType.ClassifierMixin, EstimatorType.RegressorMixin}),
                help="Available only if estimator is `Classifier` or `Regressor`",
                key="linear",
            )

    with st.container():  # predict_proba and decision_function
        c41, c42 = st.columns(2)

        with c41:  # predict_proba
            predict_proba = st.toggle(
                label=PROMPT_PREDICT_PROBA,
                disabled=(estimator_type not in {EstimatorType.ClassifierMixin, EstimatorType.OutlierMixin}),
                help=(
                    "[predict_proba](https://scikit-learn.org/dev/glossary.html#term-predict_proba): "
                    "Available only if estimator is `Classifier` or `Outlier`. "
                ),
                key="predict_proba",
            )

        with c42:  # decision_function
            decision_function = st.toggle(
                label=PROMPT_DECISION_FUNCTION,
                disabled=(estimator_type != EstimatorType.ClassifierMixin) or linear,
                help=(
                    "[decision_function](https://scikit-learn.org/dev/glossary.html#term-decision_function): "
                    "Available only if estimator is `Classifier`"
                ),
                key="decision_function",
            )

    st.write("#")  # empty space hack

    with st.container():  # forge button
        _, c52, _, c54 = st.columns([2, 1, 1, 1])

        with c52:
            forge_btn = st.button(
                label="Time to forge 🛠️",
                type="primary",
                disabled=any(
                    [
                        not name,
                        not estimator_type,
                        not required_is_valid,
                        not optional_is_valid,
                        msg_duplicated_params,
                    ]
                ),
                key="forge_btn",
            )
            if forge_btn:
                st.session_state["forge_counter"] += 1
                st.session_state["forged_template"] = render_template(
                    name=name,
                    estimator_type=estimator_type,  # type: ignore[arg-type]  # At this point estimator_type is never None.
                    required=required,
                    optional=optional,
                    linear=linear,
                    sample_weight=sample_weight,
                    predict_proba=predict_proba,
                    decision_function=decision_function,
                )

        with c54, st.popover(label="Download", disabled=not st.session_state["forge_counter"]):
            if name:
                file_name = st.text_input(label="Select filename", value=f"{name.lower()}.py", key="file_name")

                data = st.session_state["forged_template"]
                st.download_button(
                    label="Confirm",
                    type="primary",
                    data=data,
                    file_name=file_name,
                    key="download_btn",
                )

    with st.container():  # code output
        if forge_btn:
            st.toast("Request submitted!")
            progress_text = "Forging in progress ..."
            progress_bar = st.progress(0, text=progress_text)
            # Consider using status component instead
            # https://docs.streamlit.io/develop/api-reference/status/st.status

            for percent_complete in range(100):
                time.sleep(0.002)
                progress_bar.progress(percent_complete + 1, text=progress_text)

            time.sleep(0.2)
            progress_bar.empty()

        if st.session_state["forge_counter"]:
            st.code(st.session_state["forged_template"], language="python", line_numbers=True)


if __name__ == "__main__":
    app()



import sys
from importlib import metadata, resources
from typing import ClassVar

from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, ScrollableContainer
from textual.reactive import reactive
from textual.widgets import Button, Collapsible, Footer, Header, Rule, Static, TextArea

from sksmithy.tui._components import (
    DecisionFunction,
    DestinationFile,
    Estimator,
    ForgeButton,
    ForgeRow,
    Linear,
    Name,
    Optional,
    PredictProba,
    Required,
    SampleWeight,
    SaveButton,
    Sidebar,
)

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self


class ForgeTUI(App):
    """Textual app to forge scikit-learn compatible estimators."""

    CSS_PATH: ClassVar[str] = str(resources.files("sksmithy") / "_static" / "tui.tcss")
    TITLE: ClassVar[str] = "Scikit-learn Smithy ⚒️"  # type: ignore[misc]

    BINDINGS: ClassVar = [
        ("ctrl+d", "toggle_sidebar", "Description"),
        ("L", "toggle_dark", "Light/Dark mode"),
        ("F", "forge", "Forge"),
        ("ctrl+s", "save", "Save"),
        ("E", "app.quit", "Exit"),
    ]

    show_sidebar = reactive(False)  # noqa: FBT003

    def on_mount(self: Self) -> None:
        """Compose on mount.

        Q: is this needed???
        """
        self.compose()

    def compose(self: Self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Container(
            Header(icon=f"v{metadata.version('sklearn-smithy')}"),
            ScrollableContainer(
                Horizontal(Name(), Estimator()),
                Horizontal(Required(), Optional()),
                Horizontal(SampleWeight(), Linear()),
                Horizontal(PredictProba(), DecisionFunction()),
                Rule(),
                ForgeRow(
                    Static(),
                    ForgeButton(),
                    SaveButton(),
                    DestinationFile(),
                ),
                Rule(),
                Collapsible(
                    TextArea(
                        text="",
                        language="python",
                        theme="vscode_dark",
                        show_line_numbers=True,
                        tab_behavior="indent",
                        id="code-area",
                    ),
                    title="Code Editor",
                    collapsed=True,
                    id="code-editor",
                ),
            ),
            Sidebar(classes="-hidden"),
            Footer(),
        )

    def action_toggle_dark(self: Self) -> None:  # pragma: no cover
        """Toggle dark mode."""
        self.dark = not self.dark

    def action_toggle_sidebar(self: Self) -> None:  # pragma: no cover
        """Toggle sidebar component."""
        sidebar = self.query_one(Sidebar)
        self.set_focus(None)

        if sidebar.has_class("-hidden"):
            sidebar.remove_class("-hidden")
        else:
            if sidebar.query("*:focus"):
                self.screen.set_focus(None)
            sidebar.add_class("-hidden")

    def action_forge(self: Self) -> None:
        """Press forge button."""
        forge_btn = self.query_one("#forge-btn", Button)
        forge_btn.press()

    def action_save(self: Self) -> None:
        """Press save button."""
        save_btn = self.query_one("#save-btn", Button)
        save_btn.press()


if __name__ == "__main__":  # pragma: no cover
    tui = ForgeTUI()
    tui.run()



import sys
import webbrowser
from importlib import resources
from pathlib import Path

from result import Err, Ok
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Grid, Horizontal, ScrollableContainer
from textual.widgets import Button, Collapsible, Input, Markdown, Select, Static, Switch, TextArea

from sksmithy._models import EstimatorType
from sksmithy._parsers import check_duplicates, name_parser, params_parser
from sksmithy._prompts import (
    PROMPT_DECISION_FUNCTION,
    PROMPT_ESTIMATOR,
    PROMPT_LINEAR,
    PROMPT_NAME,
    PROMPT_OPTIONAL,
    PROMPT_OUTPUT,
    PROMPT_PREDICT_PROBA,
    PROMPT_REQUIRED,
    PROMPT_SAMPLE_WEIGHT,
)
from sksmithy._utils import render_template
from sksmithy.tui._validators import NameValidator, ParamsValidator

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self


SIDEBAR_MSG: str = (resources.files("sksmithy") / "_static" / "description.md").read_text()


class Prompt(Static):
    pass


class Name(Container):
    """Name input component."""

    def compose(self: Self) -> ComposeResult:
        yield Prompt(PROMPT_NAME, classes="label")
        yield Input(placeholder="MightyEstimator", id="name", validators=[NameValidator()])

    @on(Input.Changed, "#name")
    def on_input_change(self: Self, event: Input.Changed) -> None:
        if not event.validation_result.is_valid:  # type: ignore[union-attr]
            self.notify(
                message=event.validation_result.failure_descriptions[0],  # type: ignore[union-attr]
                title="Invalid Name",
                severity="error",
                timeout=5,
            )
        else:
            output_file = self.app.query_one("#output-file", Input)
            output_file.value = f"{event.value.lower()}.py"


class Estimator(Container):
    """Estimator select component."""

    def compose(self: Self) -> ComposeResult:
        yield Prompt(PROMPT_ESTIMATOR, classes="label")
        yield Select(
            options=((" ".join(x.capitalize() for x in e.value.split("-")), e.value) for e in EstimatorType),
            id="estimator",
        )

    @on(Select.Changed, "#estimator")
    def on_select_change(self: Self, event: Select.Changed) -> None:
        linear = self.app.query_one("#linear", Switch)
        predict_proba = self.app.query_one("#predict_proba", Switch)
        decision_function = self.app.query_one("#decision_function", Switch)

        linear.disabled = event.value not in {"classifier", "regressor"}
        predict_proba.disabled = event.value not in {"classifier", "outlier"}
        decision_function.disabled = event.value not in {"classifier"}

        linear.value = linear.value and (not linear.disabled)
        predict_proba.value = predict_proba.value and (not predict_proba.disabled)
        decision_function.value = decision_function.value and (not decision_function.disabled)


class Required(Container):
    """Required params input component."""

    def compose(self: Self) -> ComposeResult:
        yield Prompt(PROMPT_REQUIRED, classes="label")
        yield Input(placeholder="alpha,beta", id="required", validators=[ParamsValidator()])

    @on(Input.Submitted, "#required")
    def on_input_change(self: Self, event: Input.Submitted) -> None:
        if not event.validation_result.is_valid:  # type: ignore[union-attr]
            self.notify(
                message="\n".join(event.validation_result.failure_descriptions),  # type: ignore[union-attr]
                title="Invalid Parameter",
                severity="error",
                timeout=5,
            )

        optional = self.app.query_one("#optional", Input).value or ""
        if (
            optional
            and event.value
            and (
                duplicates_result := check_duplicates(
                    event.value.split(","),
                    optional.split(","),
                )
            )
        ):
            self.notify(
                message=duplicates_result,
                title="Duplicate Parameter",
                severity="error",
                timeout=5,
            )


class Optional(Container):
    """Optional params input component."""

    def compose(self: Self) -> ComposeResult:
        yield Prompt(PROMPT_OPTIONAL, classes="label")
        yield Input(placeholder="mu,sigma", id="optional", validators=[ParamsValidator()])

    @on(Input.Submitted, "#optional")
    def on_optional_change(self: Self, event: Input.Submitted) -> None:
        if not event.validation_result.is_valid:  # type: ignore[union-attr]
            self.notify(
                message="\n".join(event.validation_result.failure_descriptions),  # type: ignore[union-attr]
                title="Invalid Parameter",
                severity="error",
                timeout=5,
            )

        required = self.app.query_one("#required", Input).value or ""
        if (
            required
            and event.value
            and (
                duplicates_result := check_duplicates(
                    required.split(","),
                    event.value.split(","),
                )
            )
        ):
            self.notify(
                message=duplicates_result,
                title="Duplicate Parameter",
                severity="error",
                timeout=5,
            )


class SampleWeight(Container):
    """sample_weight switch component."""

    def compose(self: Self) -> ComposeResult:
        yield Horizontal(
            Switch(id="sample_weight"),
            Prompt(PROMPT_SAMPLE_WEIGHT, classes="label"),
            classes="container",
        )


class Linear(Container):
    """linear switch component."""

    def compose(self: Self) -> ComposeResult:
        yield Horizontal(
            Switch(id="linear"),
            Prompt(PROMPT_LINEAR, classes="label"),
            classes="container",
        )

    @on(Switch.Changed, "#linear")
    def on_switch_changed(self: Self, event: Switch.Changed) -> None:
        decision_function = self.app.query_one("#decision_function", Switch)
        decision_function.disabled = event.value
        decision_function.value = decision_function.value and (not decision_function.disabled)


class PredictProba(Container):
    """predict_proba switch component."""

    def compose(self: Self) -> ComposeResult:
        yield Horizontal(
            Switch(id="predict_proba"),
            Prompt(PROMPT_PREDICT_PROBA, classes="label"),
            classes="container",
        )


class DecisionFunction(Container):
    """decision_function switch component."""

    def compose(self: Self) -> ComposeResult:
        yield Horizontal(
            Switch(id="decision_function"),
            Prompt(PROMPT_DECISION_FUNCTION, classes="label"),
            classes="container",
        )


class ForgeButton(Container):
    """forge button component."""

    def compose(self: Self) -> ComposeResult:
        yield Button(label="Forge ⚒️", id="forge-btn", variant="success")

    @on(Button.Pressed, "#forge-btn")
    def on_forge(self: Self, _: Button.Pressed) -> None:  # noqa: C901
        errors = []

        name_input = self.app.query_one("#name", Input).value
        estimator = self.app.query_one("#estimator", Select).value
        required_params = self.app.query_one("#required", Input).value
        optional_params = self.app.query_one("#optional", Input).value

        sample_weight = self.app.query_one("#linear", Switch).value
        linear = self.app.query_one("#linear", Switch).value
        predict_proba = self.app.query_one("#predict_proba", Switch).value
        decision_function = self.app.query_one("#decision_function", Switch).value

        code_area = self.app.query_one("#code-area", TextArea)
        code_editor = self.app.query_one("#code-editor", Collapsible)

        match name_parser(name_input):
            case Ok(name):
                pass
            case Err(name_error_msg):
                errors.append(name_error_msg)

        match estimator:
            case str(v):
                estimator_type = EstimatorType(v)
            case Select.BLANK:
                errors.append("Estimator cannot be empty!")

        match params_parser(required_params):
            case Ok(required):
                required_is_valid = True
            case Err(required_err_msg):
                required_is_valid = False
                errors.append(required_err_msg)

        match params_parser(optional_params):
            case Ok(optional):
                optional_is_valid = True

            case Err(optional_err_msg):
                optional_is_valid = False
                errors.append(optional_err_msg)

        if required_is_valid and optional_is_valid and (msg_duplicated_params := check_duplicates(required, optional)):
            errors.append(msg_duplicated_params)

        if errors:
            self.notify(
                message="\n".join([f"- {e}" for e in errors]),
                title="Invalid inputs!",
                severity="error",
                timeout=5,
            )

        else:
            forged_template = render_template(
                name=name,
                estimator_type=estimator_type,
                required=required,
                optional=optional,
                linear=linear,
                sample_weight=sample_weight,
                predict_proba=predict_proba,
                decision_function=decision_function,
                tags=None,
            )

            code_area.text = forged_template
            code_editor.collapsed = False

            self.notify(
                message="Template forged!",
                title="Success!",
                severity="information",
                timeout=5,
            )


class SaveButton(Container):
    """forge button component."""

    def compose(self: Self) -> ComposeResult:
        yield Button(label="Save 📂", id="save-btn", variant="primary")

    @on(Button.Pressed, "#save-btn")
    def on_save(self: Self, _: Button.Pressed) -> None:
        output_file = self.app.query_one("#output-file", Input).value

        if not output_file:
            self.notify(
                message="Outfile filename cannot be empty!",
                title="Invalid filename!",
                severity="error",
                timeout=5,
            )
        else:
            destination_file = Path(output_file)
            destination_file.parent.mkdir(parents=True, exist_ok=True)

            code = self.app.query_one("#code-area", TextArea).text

            with destination_file.open(mode="w") as destination:
                destination.write(code)

            self.notify(
                message=f"Saved at {destination_file}",
                title="Success!",
                severity="information",
                timeout=5,
            )


class DestinationFile(Container):
    """Destination file input component."""

    def compose(self: Self) -> ComposeResult:
        yield Input(placeholder=PROMPT_OUTPUT, id="output-file")


class ForgeRow(Grid):
    """Row grid for forge."""


class OptionGroup(ScrollableContainer):
    pass


class Sidebar(Container):
    def compose(self: Self) -> ComposeResult:
        yield OptionGroup(Markdown(SIDEBAR_MSG))

    def on_markdown_link_clicked(self: Self, event: Markdown.LinkClicked) -> None:
        # Relevant discussion: https://github.com/Textualize/textual/discussions/3668
        webbrowser.open_new_tab(event.href)
